@@ -1,18 +1,32 @@
 /*
  * crypt.c
  *
- * encryptian module
+ * ufdbGuard is copyrighted (C) 2005-2010 by URLfilterDB with all rights reserved.
+ *
+ * Parts of the ufdbGuard daemon are based on squidGuard.
+ * This module is NOT based on squidGuard.
+ *
+ * Encryption module
+ *
+ * $Id: crypt.c,v 1.7 2010/09/09 20:55:24 root Exp root $
  */
 
+
 #include "ufdb.h"
 
 
+/* We want agressive optimisations: */
+/* These pragmas only work for GCC 4.4 and above */
+#pragma GCC optimize (3,"inline-functions")
+#pragma GCC optimize (3,"inline-limit=1500")
+
+
 void ufdbCryptInit( ufdbCrypt * uc, const unsigned char * key, unsigned int keySize )
 {
-   int i;
-   unsigned int t, u;
-   unsigned int ki;
-   unsigned int sti;
+   int             i;
+   unsigned int    t, u;
+   unsigned int    ki;
+   unsigned int    sti;
    unsigned char * state;
       
    state = uc->state;
@@ -60,12 +74,15 @@ static __inline__ unsigned int _nextByte( ufdbCrypt * uc )
 }
 
 
-void ufdbEncryptText( ufdbCrypt * uc, unsigned char * dest, 
-		      const unsigned char * src, unsigned int len )
+void ufdbEncryptText( 
+   ufdbCrypt *           uc, 
+   unsigned char *       dest, 
+   const unsigned char * src, 
+   unsigned int          len )
 {
    while (len > 0)
    {
-      *dest = *src ^ _nextByte(uc);
+      *dest = *src ^ _nextByte( uc );
       src++;
       dest++;
       len--;
