@@ -1,23 +1,20 @@
 /*
-  By accepting this notice, you agree to be bound by the following
-  agreements:
-  
-  ufdbGuard is copyrighted (C) 2005,2006,2007 by URLfilterDB with all rights reserved.
-  ufdbGuard is based on squidGuard.
-
-  squidGuard is copyrighted (C) 1998 by
-  ElTele st AS, Oslo, Norway, with all rights reserved.
-  
-  This program is free software; you can redistribute it and/or modify it
-  under the terms of the GNU General Public License (version 2) as
-  published by the Free Software Foundation.  It is distributed in the
-  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
-  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
-  PURPOSE.  See the GNU General Public License (GPL) for more details.
-  
-  You should have received a copy of the GNU General Public License
-  (GPL) along with this program.
-*/
+ * ufdbGuard is copyrighted (C) 2005-2010 by URLfilterDB with all rights reserved.
+ *
+ * Parts of the ufdbGuard daemon are based on squidGuard.
+ * squidGuard is copyrighted (C) 1998 by
+ * ElTele st AS, Oslo, Norway, with all rights reserved.
+ * 
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU General Public License (version 2) as
+ * published by the Free Software Foundation.  It is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
+ * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
+ * PURPOSE.  See the GNU General Public License (GPL) for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * (GPL) along with this program.
+ */
 
 
 #include "ufdb.h"
@@ -31,52 +28,54 @@
 
 #include "ufdblib.h"
 
-extern int sig_hup;           
 extern int sig_other;           
-extern int sig_alrm;           
 extern char ** globalArgv;           
 extern char ** globalEnvp;           
 extern struct Acl * defaultAcl;
 
-extern struct Source * Source;
-extern struct Destination * Dest;
 
 
 void sgHandlerSigHUP( int s )
 {
-  if (sig_hup == s)
-     abort();
-  sig_hup = s;
+   sig_hup = s;
 }
 
 
 void sgSignalHandler( int s )
 {
-  if (sig_other == s)
-     abort();
-  sig_other = s;
+   static int num = 0;
+
+   if (sig_other == s)
+   {
+      num++;
+      if (num >= 3)
+	 abort();
+   }
+   else
+      num = 0;
+
+   sig_other = s;
 }
 
 
-void sgAlarm( int signal )
+void sgAlarm( int s )
 {
-  sig_alrm = 1;
-  sgTimeNextEvent();
 }
 
 
-#if 0	
+#if 0
 static void printSquidInfo( struct SquidInfo * s )
 {
   fprintf( stderr, "   protocol = %s\n", s->protocol );
   fprintf( stderr, "   port     = %d\n", s->port     );
   fprintf( stderr, "   dot      = %d\n", s->dot      );
   UFDBprintRevURL( s->revUrl );
+  fprintf( stderr, "   orig     = %s\n", s->orig         );
+  fprintf( stderr, "   url2displ= %s\n", s->url2display  );
+  fprintf( stderr, "   url      = %s\n", s->url          );
+  fprintf( stderr, "   surl     = %s\n", s->surl         );
   fprintf( stderr, "   domain   = %s\n", s->domain       );
   fprintf( stderr, "   orgdomain= %s\n", s->orig_domain  );
-  fprintf( stderr, "   url      = %s\n", s->url          );
-  fprintf( stderr, "   orig     = %s\n", s->orig         );
-  fprintf( stderr, "   surl     = %s\n", s->surl         );
   fprintf( stderr, "   src      = %s\n", s->src          );
   fprintf( stderr, "   srcDom   = %s\n", s->srcDomain    );
   fprintf( stderr, "   ident    = %s\n", s->ident        );
@@ -86,9 +85,66 @@ static void printSquidInfo( struct SquidInfo * s )
 #endif
 
 
+static __inline__ int domain_has_tld( char * p )
+{
+   while (*p != '\0')
+   {
+      if (*p == '/')
+         return 0;
+      if (*p == '.')
+         return 1;
+      p++;
+   }
+   return 0;
+}
+
+
+static __inline__ int squeeze_html_char( char * p, int * hex )
+{
+   int length;
+
+   length = 0;
+   *hex = 0;
+   while (*p != '\0'  &&  isxdigit( (int) *p ))
+   {
+      int h;
+      h = (*p <= '9') ? *p - '0' : *p - 'a' + 10;
+      *hex = *hex * 16 + h;
+      p++;
+      length++;
+   }
+
+#if 0
+   ufdbLogMessage( "   squeeze_html_char hex=%02x  length=%d  *p=%c", *hex, length, *p );
+#endif
+
+   if (*p == ';')
+   {
+      if (*hex == 0)
+         return length;
+      if (*hex < 0x20)
+      {
+         if (*hex != '\t'  &&  *hex != '\n'  &&  *hex != '\r'  &&  *hex != '\f')
+	    *hex = ' ';
+      }
+      else if (*hex == 0x7f  ||  *hex >= 0xff)
+      {
+         *hex = ' ';
+      }
+      else if (*hex <= 'Z'  &&  *hex >= 'A')
+	 *hex += 'a' - 'A';
+      return length;
+   }
+   
+   /* '&#xxx' without trailing ';' is not a valid HTML character */
+   return -1;
+}
+
+
 /*
- * parsers the squidline:
+ * parse the squidline:
  * URL ip-address/fqdn ident method
+ * The URL is normalised/rewritten.
  * Squid pre2.6: http://www.sex.com 10.1.1.44/- - GET
  * Squid 2.6.x:  http://www.sex.com 10.1.1.44/- - GET -    (urlgroup added)
  */
@@ -97,9 +153,16 @@ int parseLine(
   char *             line, 
   struct SquidInfo * s )
 {
-  char * p, *d, *a, *e, *o, *field;
-  char * up, *upp;
+  char * p;
+  char * d;
+  char * a;
+  char * e;
+  char * o;
+  char * field;
+  char * upp;
   int    i; 
+  int    obfuscated;
+  int    is_ipv6;
   char * lineptr;
 
   s->revUrl = NULL;
@@ -108,8 +171,8 @@ int parseLine(
   s->port = 80;
   s->orig[0] = 
     s->url[0] = s->domain[0] = s->orig_domain[0] = s->src[0] = s->ident[0] = s->ident[1] =
-    s->protocol[0] = s->method[0] = s->urlgroup[0] = s->srcDomain[0] = s->surl[0] =  '\0';
-
+    s->protocol[0] = s->method[0] = s->urlgroup[0] = s->srcDomain[0] = s->surl[0] =  
+    s->url2display[0] = '\0';
 
 #if 0
   fprintf( stderr, "   line: %s\n", line );
@@ -122,48 +185,46 @@ int parseLine(
   /* prevent loops with regular expression matching: */
   p = strstr( field, "/URLblocked.cgi" );
   if (p != NULL)
-     *(p+15) = '\0';
+    *(p+15) = '\0';
 
-#if 1
   strcpy( s->orig, field );
-#else
-  for (d = s->orig, p = field; *p != '\0'; p++)
+  if (UFDBglobalShowURLdetails)
   {
-     /* The original URL is used to display the "no access" error and
-      * for reasons of privacy we remove everything after a ? or &.
-      * This also prevents loops in the display of an error message
-      * where a regular expression matches a part of a URL that is
-      * in the error message (i.e. the error message gets blocked
-      * and a new error message is generated).
-      */
-     if (*p == '?' || *p == '&')	/* strip '?...' and '&...' */
-     {
-	*d++ = '.';
-	*d++ = '.';
-	*d++ = '.';
-	break;
-     }
-     *d++ = *p;
+     strncpy( s->url2display, field, 1000 );
+     p = &(s->url2display[1000]);
   }
-  *d = '\0';
+  else
+  {
+     strncpy( s->url2display, field, 300 );
+     s->url2display[300] = '\0';
+     p = strchr( s->url2display, '?' );
+     if (p == NULL)
+     {
+	p = strchr( s->url2display, ';' );
+	if (p == NULL)
+	{
+	   p = strstr( s->url2display, "%3F" );
+	   if (p == NULL)
+	      p = &(s->url2display[300]);
+	}
+     }
+  }
+  strcpy( p, "...." );
+
+#if 0
+  ufdbLogMessage( "   parseLine: url2display = %s", s->url2display );
 #endif
 
-  for (p = field; *p != '\0'; p++) 	/* convert URL to lowercase chars */
+  for (p = field;  *p != '\0';  p++) 	/* convert URL to lowercase chars */
   {
-#if 0
-    /* This code is causing problems since a regular expression match for 
-     * Google's cache (http://64.233.169.104/search?q=cache:OK6uOgz2pp0J:www.xs4all.nl/)
-     * fails ! */
-    if (*p == '?' || *p == '&')		/* strip '?...' and '&...' */
-    {
-       *p = '\0';
-       break;
-    }
-#endif
-    if (*p >= 'A'  &&  *p <= 'Z')
+    if (*p <= 'Z'  &&  *p >= 'A')
       *p += 'a' - 'A';
   }
 
+  /* optimize for speed; trim very long URLs */
+  if (p - field > 510)
+     field[510] = '\0';
+
   p = strstr( field, "://" );
   if (p == NULL)			/* no protocol, defaults to http */
   {
@@ -183,98 +244,178 @@ int parseLine(
     field = p;
   }
 
-  /* do some url decoding */
+   /* skip the initial optional username:password@ part */
+   /* anticipate:   user:pw@www.domain.com		*/
+   /*               user:pw@www.domain.com/path		*/
+   /*               www.domain.com			*/
+   /*               www.domain.com/path			*/
+   i = 0;
+   while (*p != '\0')
+   {
+      if (*p == '/')
+         break;
+      if (*p == '@')
+      {
+         p++;
+	 i = 1;
+	 break;
+      }
+      p++;
+   }
+   if (i == 0)
+      p = field;
+
+  /* do some character decoding: translate '//' '/./' '&#xx' and '%xx' */
 
-  up = field;
   upp = s->url;
-  i = 0;
-  while(up[i] != '\0') 
+  while(*p != '\0') 
   {
-#if 0
-    if (up[i] == '?' || up[i] == '&')
+    if (*p == '%')			/* convert '%HH' into character */
     {
-       up[i] = '\0';
-       break;
-    }
-#endif
-    if (up[i] == '%')
-    {
-      if (isxdigit( (int) up[i+1]) && isxdigit( (int) up[i+2]))
+      if (isxdigit((int) *(p+1))  &&  isxdigit( (int) *(p+2) ))
       {
 	int    hex;
 	char   h1, h2;
 	/* note that the URL is already converted to lowercase */
-#if 0
-        h1 = up[i+1] >= 'a' ? up[i+1] - ('a' - 'A') : up[i+1];
-        h2 = up[i+2] >= 'a' ? up[i+2] - ('a' - 'A') : up[i+2];
-        hex = h1 >= 'A' ? h1 - 'A' - 10 : h1 - '0';
-        hex *= 16;
-        hex += h2 >= 'A' ? h2 - 'A' - 10 : h2 - '0';
-#else
-	h1 = up[i+1];
-	h2 = up[i+2];
+	h1 = *(p+1);
+	h2 = *(p+2);
 	hex  = (h1 <= '9') ? h1 - '0' : h1 - 'a' + 10;
         hex *= 16;
 	hex += (h2 <= '9') ? h2 - '0' : h2 - 'a' + 10;
-#endif
-	/* don't convert control characters */
-	if (hex <= 0x20)
+	/* be careful with conversion of control characters */
+	if (hex == 0)
 	{
-	   *upp++ = up[i++];
-	   *upp++ = up[i++];
-	   *upp++ = up[i];
+	   p += 3;
+	   continue;
+	}
+	else if (hex < 0x20)
+	{
+	   if (hex != '\t'  &&  hex != '\r'  &&  hex != '\n'  &&  hex != '\f')
+	   {
+	      hex = ' ';
+	   }
 	}
 	else
 	{ 
-#if 0
-	   if (hex == '?' || hex == '&')
+	   if (hex == 0x7f  ||  hex == 0xff)
 	   {
-	      up[i] = '\0';
-	      break;
+	      hex = ' ';
+	   }
+	   else if (hex <= 'Z'  &&  hex >= 'A')
+	   {
+		 hex += 'a' - 'A';
 	   }
-#endif
-	   if (hex >= 'A'  &&  hex <= 'Z')
-	      hex += 'a' - 'A';
-	   *upp++ = hex;
-	   i += 2;
 	}
+	*upp++ = hex;
+	p += 2;
       }
       else
       { 						/* an erroneous hex code, we ignore it */
-        *upp++ = up[i++];
-        *upp++ = up[i++];
-        *upp++ = up[i];
+        *upp++ = *p;					/* just copy one character */
       }
     }
+    else if (*p == '&'  &&  *(p+1) == '#')		/* convert '&#HH;' into character */
+    {
+       int  hex;
+       int  length;
+
+       length = squeeze_html_char( p+2, &hex );
+       if (length >= 0)
+       {
+	  if (hex != 0)
+	     *upp++ = hex;
+	  p += length + 2;
+       }
+       else  /* not a valid HTML character code; just copy one character */
+       {
+          *upp++ = *p;
+       }
+    }
     else
     {
-      while (up[i] == '/')
+      while (*p == '/')
       {
-         if (up[i+1] == '/')					/* substitute // by / */
-	    i++;
-	 else if (up[i+1] == '.'  && up[i+2] == '/')		/* substitute /./ by / */
-	    i += 2;
+         if (*(p+1) == '/')					/* substitute // by / */
+	    p++;
+	 else if (*(p+1) == '.'  &&  *(p+2) == '/')		/* substitute /./ by / */
+	    p += 2;
+	 else if (*(p+1) == '.'  &&  *(p+2) == '.'  &&  *(p+3) == '/')    /* substitute /xxxx/../ by / */
+	 {
+	    /* try to find the previous directory... */
+	    char * prevdir;
+	    prevdir = upp - 1;
+	    while (*prevdir != '/'  &&  prevdir > s->url)
+	       prevdir--;
+	    if (prevdir > s->url)
+	    {
+	       upp = prevdir;
+	       p += 3;
+	    }
+	    else
+	       break;
+	 }
 	 else
 	    break;
       }
-      *upp++ = up[i];
+      *upp++ = *p;
     }
-    i++;
+    p++;
   }
-  *upp++ = up[i];
+  *upp++ = *p;
   *upp = '\0';
 
 #if 0
-  fprintf( stderr, ">>>> s->url = %s\n", s->url );
+  ufdbLogMessage( "   parseLine: s->url = %s", s->url );
 #endif
 
-  d = strchr(p,'/'); 		/* find domain end */
-  e = d;
-  a = strchr(p,'@'); 		/* find auth  */
-  if (a != NULL  &&  (a < d || d == NULL)) 
-    p = a + 1;
+  p = s->url;
 
-  a = strchr(p,':'); 		/* find port */
+  is_ipv6 = 0;
+  if (*p == '[')			/* IPv6 address */
+  {
+     char * end;
+     char * oldBracket;
+
+     is_ipv6 = 1;
+     s->orig_domain[0] = '\0';
+     oldBracket = strchr( p, ']' );
+     if (oldBracket != NULL)
+        *oldBracket = '\0';
+     end = UFDBparseIPv6address( p, s->orig_domain );
+     if (oldBracket != NULL)
+        *oldBracket = ']';
+     if (end != NULL)
+     {
+        UFDBupdateURLwithNormalisedDomain( p, s->orig_domain );
+	/* uh-oh: the normalised domain is usually smaller and our pointers have moved */
+     }
+     if (end == NULL)		/* invalid IPv6 address */
+     {
+        end = oldBracket;
+        if (end == NULL)
+	{
+	   end = strchr( p, '/' );
+	   if (end == NULL)
+	   {
+	      /* I give up */
+	      d = e = a = NULL;
+	      goto parse2;
+	   }
+        }
+     }
+     s->dot = 1;			/* set 'dot' for IPv6 also */
+     d = strchr( end, '/' );
+     e = d;
+     a = strchr( end, ':' ); 		/* find port */
+  }
+  else					/* not IPv6 */
+  {
+     d = strchr( p, '/' ); 		/* find domain end */
+     e = d;
+     a = strchr( p, ':' ); 		/* find port */
+  }
+
+parse2:
   if (a != NULL  &&  (a < d || d == NULL))
   {
     s->port = 0;
@@ -288,29 +429,32 @@ int parseLine(
     e = a;
   }
 
-  /* how ugly, Squid does not send the https protocol string */
+  /* how ugly, Squid strips "https://" from the URL and inserts ":443" */
   if (s->port == 443)
      strcpy( s->protocol, "https" );
 
-  if (e == NULL)
-     strcpy( s->orig_domain, p );
-  else
+  if (!is_ipv6)
   {
-     i = e - p;
-     strncpy( s->orig_domain, p, i );
-     s->orig_domain[i] = '\0';
+     if (e == NULL)
+	strcpy( s->orig_domain, p );
+     else
+     {
+	i = e - p;
+	strncpy( s->orig_domain, p, i );
+	s->orig_domain[i] = '\0';
+     }
   }
   
   /* strip prefix www. www8. ftp. ftp5. ... */
   o = p;
-
-  if ((p[0] == 'w' && p[1] == 'w' && p[2] == 'w') ||		/* match www. */
-      (p[0] == 'f' && p[1] == 't' && p[2] == 'p'))  		/* match ftp. */
+  if ((p[0] == 'w' && p[1] == 'w' && p[2] == 'w') ||		/* match "www" */
+      (p[0] == 'f' && p[1] == 't' && p[2] == 'p'))  		/* match "ftp" */
   {
     p += 3;
-    while (p[0] <= '9' && p[0] >= '0')              		/* match www99. */
+    while (p[0] <= '9'  &&  p[0] >= '0')              		/* match www99. */
       p++;
-    if (p[0] != '.'  ||  strchr( p+1, '.' ) == NULL)		/* skip www_xxx.com and www99.com */
+    								/* skip www_xxx.com and www99.com */
+    if (*p != '.'  ||  !domain_has_tld(p+1))
       p = o;  /* not a hostname */
     else
       p++;
@@ -328,23 +472,66 @@ int parseLine(
   }
   else 
   {
-    memcpy( s->domain, p, e - p );	
-    *(s->domain + (e - p)) = '\0';
+     memcpy( s->domain, p, e - p );	
+     *(s->domain + (e - p)) = '\0';
   }
+
   strcpy( s->surl, p );
 
 #if 0
-  fprintf( stderr, ">>>>> s->surl = %s\n", s->surl );
+  ufdbLogMessage( "   parseLine: s->surl = %s", s->surl );
 #endif
 
+  /* determine if the domainname is an IP address.
+   * At the same time scan for IP address obfuscations, e.g. 066.1.2.3
+   */
+  obfuscated = (s->domain[0] == '0');
   for (o = s->domain; isdigit( (int) *o) || *o == '.'; o++)
-     ;
+  {
+     if (*o == '.'  &&  *(o+1) == '0')
+        obfuscated = 1;
+  }
   if (*o == '\0')
-    s->dot = 1;
+  {
+     s->dot = 1;
+     if (obfuscated)
+     {
+        /* rewrite s->surl to remove the obfuscation */
+	d = p = s->surl;
+	/* remove obfuscating leading zeroes */
+	while (*p == '0'  &&  *(p+1) != '.'  &&  *(p+1) != '/'  &&  *(p+1) != '\0')
+	   p++;
+        while (*p != '\0'  &&  *p != '/')
+	{
+	   if (*p == '.'  &&  *(p+1) == '0')
+	   {
+	      *d++ = '.';
+	      p++;
+	      /* remove obfuscating zeroes */
+	      while (*p == '0'  &&  *(p+1) != '.'  &&  *(p+1) != '/'  &&  *(p+1) != '\0')
+		 p++;
+	   }
+	   else
+	   {
+	      *d = *p;
+	      d++;
+	      p++;
+	   }
+	}
+	/* copy the URI */
+	while (*p != '\0')
+	{
+	   *d = *p;
+	   d++;
+	   p++;
+        }
+        *d = '\0';
+     }
+  }
 
   if ((p = strtok_r(NULL," \t\n",&lineptr)) != NULL)
   {
-    o = strchr(p,'/');
+    o = strchr( p, '/' );
     if (o != NULL) 
     {
       memcpy( s->src, p, o-p );
@@ -356,7 +543,7 @@ int parseLine(
     else
     {
       strcpy( s->src, p );
-      s->srcDomain[0] = '\0';	/* TODO: check */
+      s->srcDomain[0] = '\0';	
     }
 
     if ((p = strtok_r(NULL," \t\n",&lineptr)) != NULL)		/* parse ident */
@@ -364,32 +551,35 @@ int parseLine(
       if (p[0] != '-'  ||  p[1] != '\0')
       {
 	strcpy( s->ident, p );
-	for (p = s->ident; *p != '\0'; p++) 	/* convert ident to lowercase chars */
+	for (p = s->ident; *p != '\0'; p++) 			/* convert ident to lowercase chars */
 	{
-	  if (*p >= 'A'  &&  *p <= 'Z')
+	  if (*p <= 'Z'  &&  *p >= 'A')
 	    *p += 'a' - 'A';
         }
       }
 
       if ((p = strtok_r(NULL," \t\n",&lineptr)) != NULL)	/* parse method, e.g. GET/CONNECT */
+      {
          strcpy( s->method, p );
-
-      if ((p = strtok_r(NULL," \t\n",&lineptr)) != NULL)	/* parse optional urlgroup (new in squid 2.6) */
-	 strcpy( s->urlgroup, p );
+	 if ((p = strtok_r(NULL," \t\n",&lineptr)) != NULL)	/* parse optional urlgroup (new in squid 2.6) */
+	    strcpy( s->urlgroup, p );
+	 else
+	    strcpy( s->urlgroup, "#" );
+      }
       else
-         strcpy( s->urlgroup, "#" );
+	 strcpy( s->urlgroup, "#" );
     }
   }
+  else
+  {
+     strcpy( s->urlgroup, "#" );
+  }
 
   if (s->domain[0] == '\0')
     return 0;
   if (s->method[0] == '\0')
     return 0;
 
-#if 0
-  fprintf( stderr, ">>>>>> s->surl = %s\n", s->surl );
-#endif
-
   /* s->revUrl is pointing to malloced memory and must be freed at the end of the main loop ! */
   s->revUrl = UFDBgenRevURL( admin, (unsigned char *) s->surl );
 
@@ -401,39 +591,16 @@ int parseLine(
 }
 
 
-/* 
- *  returns a pointer to the domain part of a fully-qualified  hostname
- *  so www.abc.xyz.dom/index.html -> xyz.dom/index.html
- */
-char * sgSkipHostPart( char * domain )
-{
-  char *p = domain , *d1 = NULL, *d2 = NULL, *path = NULL;
-
-  if ((path = (char *) strchr(p,'/')) == NULL)
-    path = domain; 
-
-  while((p = (char *) strchr(p,'.')) != NULL )
-  {
-    if (p > path && path != domain)
-      break;
-    d2 = d1;
-    d1 = p;
-    p++;
-  }
-  if (d2 != NULL)
-    return d2+1;
-
-  return domain;
-}
-
 /*
  * checks the vality of a dotted address. 
  */
-ulong * sgConvDot( char * dot, ulong * ipaddr )
+ulong * sgConvDot( 
+   char *  dot, 
+   ulong * ipaddr )
 {
-   char * t;
-   int    octet;
-   int    shift;
+   char *  t;
+   int     octet;
+   int     shift;
 
    *ipaddr = 0;
    shift = 24;
@@ -469,85 +636,10 @@ ulong * sgConvDot( char * dot, ulong * ipaddr )
 }
 
 
-/*
- * Reverses cmp of strings
- */
-int sgStrRcmp( char * a, char * b )
-{
-  char *a1 = (char *) strchr(a,'\0');
-  char *b1 = (char *) strchr(b,'\0');
-
-  while(*a1 == *b1)
-  {
-    if (b1 == b || a1 == a)
-      break;
-    a1--; b1--;
-  }
-  if (a1 == a && b1 == b)
-    return *a1 - *b1; 
-  if (a1 == a)
-    return -1;
-  if (b1 == b)
-    return 1;
-
-  return *a1 - *b1;
-}
-
-
-int sgStrRncmp( char * a, char * b, int blen )
-{
-  char *a1 = (char *) strchr(a,'\0');
-  char *b1 = (char *) strchr(b,'\0');
-
-  while(*a1 == *b1 && blen > 0 )
-  {
-    if (b1 == b || a1 == a)
-      break;
-    a1--; b1--; blen--;
-  }
-
-  if (a1 == a && b1 == b)
-    return *a1 - *b1; 
-
-  if (blen == 0)
-    return *a1 - *b1;
-
-  if (a1 == a)
-    return -1;
-
-  if (b1 == b)
-    return 1;
-
-  return *a1 - *b1;
-}
-
-
-/*
- *  sgDomStrRncmp checks if p2 is equal to or a subdomain of p1
- */
-int sgDomStrRcmp( char * p1, char * p2 )
-{
-  char *p1e = (char *) strchr(p1,'\0');
-  char *p2e = (char *) strchr(p2,'\0');
-
-  for(;p1e >= p1 && p2e >= p2 && *p1e == *p2e; p1e--, p2e--)
-     ;
-
-  if (p1e < p1 && p2e < p2)
-    return 0;
-  if (p2e < p2)
-    return -*p1e;
-  if (p1e < p1 && *p2e == '.')	/* found a subdomain */
-    return 0;
-  return *p1e - *p2e;
-}
-
-
 /*
  *  Regexp functions
  */
 
-
 char * ufdbRegExpSubst( struct ufdbRegExp * regexp, char * pattern, char * newstring )
 {
   struct ufdbRegExp * re;
@@ -572,15 +664,16 @@ char * ufdbRegExpSubst( struct ufdbRegExp * regexp, char * pattern, char * newst
 	*newstring = '\0';
       p = newstring;
       do {
-	if ((p - newstring) + pm.rm_so >= MAX_BUF)
+	if ((p - newstring) + pm.rm_so >= UFDB_MAX_URL_LENGTH)
 	  break;
 	p = strncat( newstring, pattern, pm.rm_so );
-	if ((p - newstring) + substlen >= MAX_BUF)
+	if ((p - newstring) + substlen >= UFDB_MAX_URL_LENGTH)
 	  break;
 	p = strcat( newstring, re->substitute );	
 	pattern = pattern + pm.rm_eo;
       } while (UFDBregexec(re->compiled, pattern, 1, &pm, REG_NOTBOL) == 0 && re->global);
-      if ((p - newstring) + strlen(pattern) <= MAX_BUF)
+
+      if ((p - newstring) + strlen(pattern) <= UFDB_MAX_URL_LENGTH)
 	p = strcat( newstring, pattern );
       result = newstring;
       break;
@@ -616,33 +709,24 @@ char * sgParseRedirect(
   char * t;
   char * d;
 
-#ifdef UFDB_USERQUOTA_SUPPORT
+#if UFDB_USERQUOTA_SUPPORT
   struct Source * s = req->lastActiveSource;
 #endif
 
   if (aclpass == NULL)
     aclpass = defaultAcl->pass;
 
+  q = p = redirect;
   buf[0] = '\0';
-  q = NULL;
-  p = redirect;
 
-  if (force_302  &&  strcmp(redirect,"302:") != 0)
+  if (force_302  &&  strncmp(redirect,"302:",4) != 0)
   {
      strcpy( buf, "302:" );
   }
 
-  while((p = strchr(p,'%')) != NULL)
+  while ((p = strchr(p,'%')) != NULL)
   {
-    if (q == NULL) 
-    {
-      memcpy( buf, redirect, p - redirect );
-      buf[p - redirect] = '\0';
-    }
-    else 
-    {
-      strncat( buf, q,  p - q );
-    }
+    strncat( buf, q,  p - q );
 
     p++;
     switch (*p)
@@ -676,7 +760,7 @@ char * sgParseRedirect(
        break;
 
     case 'i': 			/* Source User Ident */
-      if ((req->ident[0] == '\0') || (req->ident[0] == '-'  &&  req->ident[1] == '\0'))
+      if ((req->ident[0] == '\0')  ||  (req->ident[0] == '-' && req->ident[1] == '\0'))
       {
 	strcat( buf, "unknown" );
       }
@@ -687,7 +771,7 @@ char * sgParseRedirect(
       break;
 
     case 'q': 				/* userquota info */
-#ifdef UFDB_USERQUOTA_SUPPORT
+#if UFDB_USERQUOTA_SUPPORT
       if (s != NULL  && 
           s->userquota.seconds != 0  &&  
 	  strcmp(req->ident, "-") != 0) 
@@ -716,7 +800,7 @@ char * sgParseRedirect(
       break;
 
     case 'n': 		/* Source Domain Name */
-      if ((req->srcDomain[0] == '\0') || (req->srcDomain[0] == '-'  &&  req->srcDomain[1] == '\0'))
+      if ((req->srcDomain[0] == '\0')  ||  (req->srcDomain[0] == '-' && req->srcDomain[1] == '\0'))
       {
 	strcat(buf, "unknown" );
       }
@@ -727,9 +811,9 @@ char * sgParseRedirect(
       break;
 
     case 'p': 		/* The url path */
-      if ((t = strstr(req->orig,"//")) != NULL)
+      if ((t = strstr(req->orig,"://")) != NULL)
       {
-        t += 2;
+        t += 3;
         if ((t = strchr(t,'/')) != NULL)
 	{
 	  t++;
@@ -747,10 +831,41 @@ char * sgParseRedirect(
       break;
 
     case 's': 		/* Source Class Matched */
-      if (acl->source == NULL || acl->source->name == NULL)
+      if (acl->source == NULL  ||  acl->source->name == NULL)
 	strcat( buf, "default" );
       else
-	strcat( buf, acl->source->name );
+      {
+	if (acl->within == UFDB_ACL_NONE)
+	   strcat( buf, acl->name );
+	else
+	{
+	   struct Acl * acl0;
+
+	   acl0 = acl;
+	   strcat( buf, acl0->name );
+	   if (acl0->within == UFDB_ACL_ELSE)
+	   {
+	      strcat( buf, "-not" );  
+	      acl0 = sgAclFindName( acl0->name );
+	      if (acl0 == NULL)
+	      {
+	         strcat( buf, "-ERROR-find-acl" );
+		 acl0 = acl;
+	      }
+	   }
+	   switch (acl0->within)
+	   {
+	   case UFDB_ACL_WITHIN:     
+	      strcat( buf, "-within-" );  
+	      strcat( buf, acl0->time->name );
+	      break;
+	   case UFDB_ACL_OUTSIDE:     
+	      strcat( buf, "-outside-" );  
+	      strcat( buf, acl0->time->name );
+	      break;
+	   }
+	}
+      }
       break;
 
     case 't': 		/* Target Class Matched */
@@ -768,38 +883,61 @@ char * sgParseRedirect(
       break;
 
     case 'u': 		/* Requested URL */
-      strcat( buf, req->orig );
+      					/* TODO: strcat( buf, req->url2display ); */
+      d = buf + strlen( buf );
+      t = req->orig;
+      while (*t != '\0')
+      {
+         if (*t == '&'  ||  *t == '?'  ||  *t == ';')
+	 {
+	    *d++ = *t;
+	    *d++ = '.';
+	    *d++ = '.';
+	    *d++ = '.';
+	    break;
+	 }
+         *d++ = *t++;
+      }
+      *d = '\0';
+      break;
+
+    case 'U': 		/* parsed URL */
+      d = buf + strlen( buf );
+      t = req->surl;
+      while (*t != '\0')
+      {
+         if (*t == '&'  ||  *t == '?'  ||  *t == ';')
+	 {
+	    *d++ = *t;
+	    *d++ = '.';
+	    *d++ = '.';
+	    *d++ = '.';
+	    break;
+	 }
+         *d++ = *t++;
+      }
+      *d = '\0';
       break;
 
     default:
       strcat( buf, "%" );
-
     }
     p++;
     q = p;
   }
 
-  if (buf[0] == '\0')
-    q = redirect;
-  else 
-    q = buf;
-
-  return q;
-}
-
-
-void sgEmergency( void )
-{
-  char buf[MAX_BUF];
-
-  ufdbLogError("going into emergency mode");
-  while(fgets(buf, MAX_BUF, stdin) != NULL)
+  if (q == redirect)   	/* no '%' found in redirect string */
   {
-    puts("");
-    fflush(stdout);
+    if (force_302  &&  strncmp( redirect, "302:", 4 ) != 0)
+    {
+       strcpy( buf, "302:" );
+       strcat( buf, redirect );
+    }
+    else
+       strcpy( buf, redirect );
   }
-  ufdbLogError("ending emergency mode, stdin empty");
-  exit(1);
+
+  return buf;
 }
 
 
@@ -857,7 +995,7 @@ struct UserQuotaInfo * setuserquota( struct UserQuotaInfo * uq )
   uq->time = 0; 
   uq->last = 0; 
   uq->consumed = 0; 
-  uq->status = 0; 
+  uq->status = UDFB_Q_STAT_FIRSTTIME; 
 
   return uq;
 }
