@@ -1,13 +1,12 @@
 /*
  * ufdbchkport.h - URLfilterDB
  *
- * ufdbGuard is copyrighted (C) 2005,2006,2007 by URLfilterDB with all rights reserved.
- * ufdbGuard is based on squidGuard.
+ * ufdbGuard is copyrighted (C) 2005-2009 by URLfilterDB with all rights reserved.
+ * 
+ * Parts of the ufdbGuard daemon are based on squidGuard.
+ * This module is NOT based on squidGuard.
  *
- * RCS $Id: ufdbchkport.h,v 1.3 2007/11/13 16:44:08 root Exp root $
- */
-
-/* 
+ * RCS $Id: ufdbchkport.h,v 1.5 2010/05/10 17:15:55 root Exp root $
  */
 
 #ifndef UFDB_UFDBCHKPORT_H_INCLUDED
@@ -18,9 +17,7 @@
 
 int UFDBloadAllowedHTTPSsites( char * filename );
 
-extern int httpsChecks;
-
-#define UFDBgetTunnelCheckMethod()		httpsChecks
+#define UFDBgetTunnelCheckMethod()		UFDBglobalTunnelCheckMethod
 void UFDBsetTunnelCheckMethod( int method );
 
 int UFDBinitHTTPSchecker( void );
